# -------------------------------------------------------------------------------------------------
# test_benchmark_mp_xi2.py: benchmarks for the multiprocessing package                              
#
# Theory:      Multiprocessing
#
# Description: Multiprocessing should be faster than threading
#              the current threading system in Python does not support multiprocssing
#              the speed fo multtiprocessing depends (among other things) on the number of cores embedded inside CPU
#
# class multiprocessing.Process([group[, target[, name[, args[, kwargs]]]]]) 
# 
#    Process objects represent activity that is run in a separate process. 
#    The Process class has equivalents of all the methods of threading.Thread.
#    
#constructor
#    should always be called with keyword arguments. 
#    
#group
#    should always be None; it exists solely for compatibility with Thread. 
#    
#target
#    is the callable object to be invoked by the run() method. 
#    It defaults to None, meaning nothing is called. 
#    
#name
#    is the process name
#    by default, a unique name is constructed of the form Process-N1:N2:...:Nk
#    where N1,N2,...,Nk is a sequence of integers whose length is determined by the generation of the process. 
#    
#args
#    is the argument tuple for the target invocation
#    
#kwargs
#    is a dictionary of keyword arguments for the target invocation. By default, no arguments are passed to target.
#    
#If a subclass overrides the constructor, it must make sure it invokes the base class constructor (Process.__init__()) before doing anything else to the process.
#    
#run() 
#    Method representing the process activity.
#    You may override this method in a subclass. 
#    The standard run() method invokes the callable object passed to the object constructor as the target argument, if any, 
#    with sequential and keyword arguments taken from the args and kwargs arguments, respectively.
#    
#start() 
#    Start the process activity.
#    This must be called at most once per process object. 
#    It arranges for the object run() method to be invoked in a separate process.
#    
#join([timeout]) 
#    Block the calling thread until the process whose join() method is called terminates or until the optional timeout occurs.
#    If timeout is None then there is no timeout.
#    A process can be joined many times.
#    A process cannot join itself because this would cause a deadlock. It is an error to attempt to join a process before it has been started.
#    
#name 
#    Return the process name.
#    
#Process.name = name 
#    Set the process name.
#    The name is a string used for identification purposes only. 
#    It has no semantics. 
#    Multiple processes may be given the same name. 
#    The initial name is set by the constructor.
#    
#is_alive() <--- OK
#    Return whether the process is alive.
#    Roughly, a process object is alive from the moment the start() method returns until the child process terminates.
#    
#daemon 
#    Return the process daemon flag., this is a boolean.
#    
#Process.daemon = daemonic 
#    Set the process daemon flag to the Boolean value daemonic. 
#    This must be called before start() is called.
#    The initial value is inherited from the creating process.
#    When a process exits, it attempts to terminate all of its daemonic child processes.
#    Note that a daemonic process is not allowed to create child processes. 
#    Otherwise a daemonic process would leave its children orphaned if it gets terminated when its parent process exits.
#        
#In addition process objects also support the following methods:
#    
#pid 
#    Return the process ID. Before the process is spawned, this will be None.
#    
#exitcode 
#    Return the child exit code. This will be None if the process has not yet terminated. A negative value -N indicates that the child was terminated by signal N.
#
#authkey 
#    Return the process authentication key (a byte string).
#    When multiprocessing is initialized the main process is assigned a random string using os.random().
#    When a Process object is created, it will inherit the authentication key of its parent process, although this may be changed using Process.authkey below.
#    See Authentication keys.
# 
#Process.authkey = authkey 
# Set the process authentication key which must be a byte string.
# 
#terminate() 
#    Terminate the process. 
#    On Unix this is done using the SIGTERM signal; 
#    on Windows TerminateProcess is used. Note that exit handlers and finally clauses, etc., will not be executed.
#    Note that descendant processes of the process will not be terminated, they will simply become orphaned
#    Warning
#    If this method is used when the associated process is using a pipe or queue 
#    then the pipe or queue is liable to become corrupted and may become unusable by other process. 
#    Similarly, if the process has acquired a lock or semaphore etc. then terminating it is liable to cause other processes to deadlock.
#    Note that the start(), join(), is_alive() and get_exit_code() methods should only be called by the process that created the process object.
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Imports from standard modules                              
# -------------------------------------------------------------------------------------------------
import os
import sys
from ctypes import Structure, c_double
import time
import threading
import gc
import random

import queue # <--- NOT Queue as in the original example

import operator

# -------------------------------------------------------------------------------------------------
# Imports from multiprocessing package                              
# -------------------------------------------------------------------------------------------------
import multiprocessing

from multiprocessing import Process
from multiprocessing import Queue
from multiprocessing import Pipe
from multiprocessing import Lock
from multiprocessing import Manager
from multiprocessing import Pool
from multiprocessing import current_process
from multiprocessing import freeze_support


from multiprocessing.sharedctypes import Value, Array

from multiprocessing.managers import BaseManager, BaseProxy


# -------------------------------------------------------------------------------------------------
# yourDragonXi imports                               
# -------------------------------------------------------------------------------------------------
from process2xi import Process2Xi
from queue2xi  import Queue2Xi
from pipe2xi import Pipe2Xi
from lock2xi import Lock2Xi
from rlock2xi import RLock2Xi
from manager2xi import Manager2Xi 
from pool2xi import Pool2Xi
from value2xi import Value2Xi
from arrayfunction2xi import ArrayFunction2Xi #<--- note: this must be called (Array2Xi uses the class from managers.py)
from basemanager2xi import BaseManager2Xi
from baseproxy2xi import BaseProxy2Xi
from condition2xi import Condition2Xi  


# -------------------------------------------------------------------------------------------------
if sys.platform == 'win32':
    _timer = time.clock
else:
    _timer = time.time

delta = 1

# -------------------------------------------------------------------------------------------------
# queuespeed_func                              
# -------------------------------------------------------------------------------------------------
def queuespeed_func(q, c, iterations):
    a = '0' * 256
    c.acquire()
    c.notify()
    c.release()

    for i in range(iterations):
        q.put(a)

    q.put('STOP')
# -------------------------------------------------------------------------------------------------
# test_queuespeed                              
# -------------------------------------------------------------------------------------------------
def test_queuespeed(Process, q, c):
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        p = Process(target=queuespeed_func, args=(q, c, iterations))
        c.acquire()
        p.start()
        c.wait()
        c.release()

        result = None
        t = _timer()

        while result != 'STOP':
            result = q.get()

        elapsed = _timer() - t

        p.join()

    print (iterations, 'objects passed through the queue in', elapsed, 'seconds')
    print ('average number/sec:', iterations/elapsed)

# -------------------------------------------------------------------------------------------------
# pipe_func                              
# -------------------------------------------------------------------------------------------------
def pipe_func(c, cond, iterations):
    a = '0' * 256
    cond.acquire()
    cond.notify()
    cond.release()

    for i in range(iterations):
        c.send(a)

    c.send('STOP')
# -------------------------------------------------------------------------------------------------
# test_pipespeed                              
# -------------------------------------------------------------------------------------------------
def test_pipespeed():
    c, d = multiprocessing.Pipe()
    cond = multiprocessing.Condition()
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        p = multiprocessing.Process(target=pipe_func,
                                    args=(d, cond, iterations))
        cond.acquire()
        p.start()
        cond.wait()
        cond.release()

        result = None
        t = _timer()

        while result != 'STOP':
            result = c.recv()

        elapsed = _timer() - t
        p.join()

    print (iterations, 'objects passed through connection in',elapsed,'seconds')
    print ('average number/sec:', iterations/elapsed)

    
# -------------------------------------------------------------------------------------------------
# test_pipespeed2xi                              
# -------------------------------------------------------------------------------------------------
def test_pipespeed2xi():
    c, d = Pipe2Xi()
    cond = Condition2Xi()
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        p = Process2Xi(target=pipe_func, args=(d, cond, iterations))
        cond.acquire()
        p.start()
        cond.wait()
        cond.release()

        result = None
        t = _timer()

        while result != 'STOP':
            result = c.recv()

        elapsed = _timer() - t
        p.join()

    print (iterations, 'objects passed through connection in',elapsed,'seconds')
    print ('average number/sec:', iterations/elapsed)
    
    
# -------------------------------------------------------------------------------------------------
# test_seqspeed                              
# -------------------------------------------------------------------------------------------------
def test_seqspeed(seq):
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        t = _timer()

        for i in range(iterations):
            a = seq[5]

        elapsed = _timer()-t

    print (iterations, 'iterations in', elapsed, 'seconds')
    print ('average number/sec:', iterations/elapsed)

# -------------------------------------------------------------------------------------------------
# test_lockspeed                              
# -------------------------------------------------------------------------------------------------
def test_lockspeed(l):
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        t = _timer()

        for i in range(iterations):
            l.acquire()
            l.release()

        elapsed = _timer()-t

    print (iterations, 'iterations in', elapsed, 'seconds')
    print ('average number/sec:', iterations/elapsed)

# -------------------------------------------------------------------------------------------------
# conditionspeed_func                              
# -------------------------------------------------------------------------------------------------
def conditionspeed_func(c, N):
    c.acquire()
    c.notify()

    for i in range(N):
        c.wait()
        c.notify()

    c.release()
# -------------------------------------------------------------------------------------------------
# test_conditionspeed                              
# -------------------------------------------------------------------------------------------------
def test_conditionspeed(Process, c):
    elapsed = 0
    iterations = 1

    while elapsed < delta:
        iterations *= 2

        c.acquire()
        p = Process(target=conditionspeed_func, args=(c, iterations))
        p.start()

        c.wait()

        t = _timer()

        for i in range(iterations):
            c.notify()
            c.wait()

        elapsed = _timer()-t

        c.release()
        p.join()

    print (iterations * 2, 'waits in', elapsed, 'seconds')
    print ('average number/sec:', iterations * 2 / elapsed)

# -------------------------------------------------------------------------------------------------
# test1: benchmart test with threading and standard multiprocessing                              
# -------------------------------------------------------------------------------------------------
def test1():
    manager = multiprocessing.Manager()

    gc.disable()

#   print ('\n\t######## testing Queue.Queue\n')
#   test_queuespeed(threading.Thread, Queue.Queue(),threading.Condition()) <--- failed
#   test_queuespeed(threading.Thread, Queue.Queue(),threading.Condition())

    print ('\n\t######## testing queue.Queue\n')  # <--- OK
    test_queuespeed(threading.Thread, queue.Queue(),threading.Condition())
    test_queuespeed(threading.Thread, queue.Queue(),threading.Condition())

    print ('\n\t######## testing multiprocessing.Queue\n')
    test_queuespeed(multiprocessing.Process, multiprocessing.Queue(),
                    multiprocessing.Condition())
    print ('\n\t######## testing Queue managed by server process\n')
    test_queuespeed(multiprocessing.Process, manager.Queue(),
                    manager.Condition())
    print ('\n\t######## testing multiprocessing.Pipe\n')
    test_pipespeed()

    print ('\n\t######## testing list\n')
    test_seqspeed(range(10))
    print ('\n\t######## testing list managed by server process\n')
    test_seqspeed(manager.list(range(10)))
    print ('\n\t######## testing Array("i", ..., lock=False)\n')

#   test_seqspeed(multiprocessing.Array('i', range(10), lock=False))  <-- failed to assetion error !
#   print ('\n\t######## testing Array("i", ..., lock=True)\n')  <-- failed to assetion error !
#   test_seqspeed(multiprocessing.Array('i', range(10), lock=True))


    
    print ('\n\t######## testing threading.Lock\n')
    test_lockspeed(threading.Lock())
    print ('\n\t######## testing threading.RLock\n')
    test_lockspeed(threading.RLock())
    print ('\n\t######## testing multiprocessing.Lock\n')
    test_lockspeed(multiprocessing.Lock())
    print ('\n\t######## testing multiprocessing.RLock\n')
    test_lockspeed(multiprocessing.RLock())
    print ('\n\t######## testing lock managed by server process\n')
    test_lockspeed(manager.Lock())
    print ('\n\t######## testing rlock managed by server process\n')
    test_lockspeed(manager.RLock())

    
    print ('\n\t######## testing threading.Condition\n')
    test_conditionspeed(threading.Thread, threading.Condition())
    print ('\n\t######## testing multiprocessing.Condition\n')
    test_conditionspeed(multiprocessing.Process, multiprocessing.Condition())
    print ('\n\t######## testing condition managed by a server process\n')
    test_conditionspeed(multiprocessing.Process, manager.Condition())

    gc.enable()

# -------------------------------------------------------------------------------------------------
# test2: benchmark test with threading and dragonized multiprocessing                              
# -------------------------------------------------------------------------------------------------
def test2():

    manager = Manager()
    manager2xi = Manager2Xi()

    gc.disable()

    print ('\n\t######## 1) testing queue speed with threading \n')  
    test_queuespeed(threading.Thread, queue.Queue(),threading.Condition())
    
    print ('\n\t######## 2) testing  queue speed with multiprocessing \n')
    test_queuespeed(multiprocessing.Process, manager.Queue(), manager.Condition())
    
    print ('\n\t######## 3) testing queue speed with dragonized multiprocessing \n')
    test_queuespeed(Process2Xi, Queue2Xi(), Condition2Xi())
    
                  
                    
    print ('\n\t######## 2) testing multiprocessing.Pipe speed \n')
    test_pipespeed()
    
    print ('\n\t######## 3) testing Pipe2Xi speed \n')
    test_pipespeed2xi()

    print ('\n\t######## 1) testing list\ speed n')
    test_seqspeed(range(10))
    
    print ('\n\t######## 2) testing manager.list speed \n')
    test_seqspeed(manager.list(range(10)))
    
    print ('\n\t######## 3) testing managerxi.list speed \n')
    test_seqspeed(manager2xi.list(range(10)))
    
    
    print ('\n\t######## 1) testing Array("i", ..., lock=False)  speed  \n')

#   test_seqspeed(multiprocessing.Array('i', range(10), lock=False))  <-- failed to assetion error !
#   print ('\n\t######## testing Array("i", ..., lock=True)\n')  <-- failed to assetion error !
#   test_seqspeed(multiprocessing.Array('i', range(10), lock=True))


    print ('\n\t######## 1) testing threading.Lock speed \n')
    test_lockspeed(threading.Lock())
    print ('\n\t######## 2) testing multiprocessing.Lock speed \n')
    test_lockspeed(multiprocessing.Lock())
    print ('\n\t######## 3) testing Lock2Xi speed \n')
    test_lockspeed(Lock2Xi())
    
        
    print ('\n\t######## 1) testing threading.RLock speed \n')
    test_lockspeed(threading.RLock())
    print ('\n\t######## 2) testing multiprocessing.RLock speed \n')
    test_lockspeed(multiprocessing.RLock())
    print ('\n\t######## 3) testing RLock2Xi speed \n')
    test_lockspeed(RLock2Xi())
    
 
    print ('\n\t######## 2) testing lock managed by server process speed\n')
    test_lockspeed(manager.Lock())
    print ('\n\t######## 3) testing RLock2Xi speed \n')
    test_lockspeed(RLock2Xi())

    print ('\n\t######## 1) testing threading.Condition speed \n')
    test_conditionspeed(threading.Thread, threading.Condition())
    print ('\n\t######## testing multiprocessing.Condition speed \n')
    test_conditionspeed(multiprocessing.Process, multiprocessing.Condition())
    print ('\n\t######## 3) Process2Xi, Condition2Xi speed \n')
    test_conditionspeed(Process2Xi, Condition2Xi())

    gc.enable()    
    
    
# -------------------------------------------------------------------------------------------------
#                               
# -------------------------------------------------------------------------------------------------
if __name__ == '__main__':
    multiprocessing.freeze_support()
    
    test1()
    
    test2()
    

# -------------------------------------------------------------------------------------------------
# phase2 Benckmark Outputs with dragonzied multiprocessing                                
# -------------------------------------------------------------------------------------------------
# 	######## 1) testing queue speed with threading  <--- OK
# 
# 32768 objects passed through the queue in 1.0732916 seconds
# average number/sec: 30530.3796284
# 
# 	######## 2) testing  queue speed with multiprocessing 
# 
# 8192 objects passed through the queue in 1.45463636 seconds
# average number/sec: 5631.64803608 <--- 82 % worse queue speed in standard multiprocessing than in threading !
# 
# 	######## 3) testing queue speed with dragonized multiprocessing 
# 
# 32768 objects passed through the queue in 1.03155716 seconds
# average number/sec: 31765.568861 <--- 4 % better dragonized queue speed than in threading, much (464%) better than multiprocessing!
# 
# -------------------------------------------------------------------------------------------------
# 	######## 2) testing multiprocessing.Pipe speed <--- OK
# 
# 131072 objects passed through connection in 1.9022642 seconds
# average number/sec: 68903.1523592
# 
# 	######## 3) testing Pipe2Xi speed 
# 
# 131072 objects passed through connection in 1.90589312 seconds
# average number/sec: 68771.9571599 <--- about the same dragonized pipe speed as in multiprocessing
# 
# -------------------------------------------------------------------------------------------------
# 	######## 1) testing list\ speed n <--- OK but avoid massive list operations in both
# 2097152 iterations in 1.65647464 seconds
# average number/sec: 1266033.26689
# 
# 	######## 2) testing manager.list speed 
# 
# 16384 iterations in 1.33876516 seconds
# average number/sec: 12238.1433948 <--- list iterations are considerably slower with multiprocessing
# 
# 	######## 3) testing managerxi.list speed 
# 
# 16384 iterations in 1.31051304 seconds
# average number/sec: 12501.9740361 <--- 2 % better with dragonized list than multiprocessing
# 
# -------------------------------------------------------------------------------------------------
# 	######## 1) testing Array("i", ..., lock=False)  speed  
# 
#
# ------------------------------------------------------------------------------------------------- 
# 	######## 1) testing threading.Lock speed  <--- OK compared to standard multiprocessing
# 
# 2097152 iterations in 1.21020716 seconds
# average number/sec: 1732886.7894
# 
# 	######## 2) testing multiprocessing.Lock speed 
# 
# 1048576 iterations in 1.27587272 seconds
# average number/sec: 821850.003972
# 
# 	######## 3) testing Lock2Xi speed 
# 
# 1048576 iterations in 1.26181304 seconds
# average number/sec: 831007.420878 <--- about the same with dragonized modules as with original multiprocessing
#
# ------------------------------------------------------------------------------------------------- 
# 	######## 1) testing threading.RLock speed <--- OK compared to standard multiprocessing
# 
# 262144 iterations in 1.41870656 seconds
# average number/sec: 184776.758909
# 
# 	######## 2) testing multiprocessing.RLock speed 
# 
# 1048576 iterations in 1.2858002 seconds
# average number/sec: 815504.61728
# 
# 	######## 3) testing RLock2Xi speed 
# 
# 1048576 iterations in 1.28755932 seconds
# average number/sec: 814390.439114 <--- about the same with dragonized modules as with original multiprocessing
#
# ------------------------------------------------------------------------------------------------- 
# 	######## 2) testing lock managed by server process speed <--- OK, even faster
# 
# 8192 iterations in 1.2973798 seconds
# average number/sec: 6314.26510572
# 
# 	######## 3) testing RLock2Xi speed 
# 
# 1048576 iterations in 1.18566864 seconds
# average number/sec: 884375.25007 <--- much faster Rlock2Xi than standard multiprocessing rlock
# 
# -------------------------------------------------------------------------------------------------
# 	######## 1) testing threading.Condition speed <--- OK compared to standard multiprocessing
# 
# 32768 waits in 1.61073232 seconds
# average number/sec: 20343.5416258
# 
# 	######## testing multiprocessing.Condition speed 
# 
# 65536 waits in 1.93772424 seconds
# average number/sec: 33821.117911
# 
# 	######## 3) Process2Xi, Condition2Xi speed 
# 
# 32768 waits in 1.01451544 seconds
# average number/sec: 32299.1634312 <--- 4.5% less in dragonized condition than in standard multiprocessing condition
# 
 
# -------------------------------------------------------------------------------------------------
# phase1 Benckmark Outputs with standard multiprocessing                                
# -------------------------------------------------------------------------------------------------
# 	######## testing queue.Queue
# 
# 32768 objects passed through the queue in 1.06565104 seconds
# average number/sec: 30749.277925
# 32768 objects passed through the queue in 1.06432692 seconds
# average number/sec: 30787.5328381
# 
# 	######## testing multiprocessing.Queue
# 
# 65536 objects passed through the queue in 1.89125756 seconds
# average number/sec: 34652.0756274
# 
# 	######## testing Queue managed by server process
# 
# 8192 objects passed through the queue in 1.33039968 seconds
# average number/sec: 6157.5480836
# 
# 	######## testing multiprocessing.Pipe
# 
# 131072 objects passed through connection in 1.94867456 seconds
# average number/sec: 67262.129188
# 
# 	######## testing list
# 
# 2097152 iterations in 1.6722324 seconds
# average number/sec: 1254103.19762
# 
# 	######## testing list managed by server process
# 
# 16384 iterations in 1.29645568 seconds
# average number/sec: 12637.53189
# 
# 	######## testing Array("i", ..., lock=False)
# 
# 
# 	######## testing threading.Lock
# 
# 2097152 iterations in 1.21092368 seconds
# average number/sec: 1731861.41673
# 
# 	######## testing threading.RLock
# 
# 262144 iterations in 1.38759256 seconds
# average number/sec: 188920.009776
# 
# 	######## testing multiprocessing.Lock
# 
# 1048576 iterations in 1.17531388 seconds
# average number/sec: 892166.780163
# 
# 	######## testing multiprocessing.RLock
# 
# 1048576 iterations in 1.19416308 seconds
# average number/sec: 878084.423779
# 
# 	######## testing lock managed by server process
# 
# 8192 iterations in 1.2613778 seconds
# average number/sec: 6494.48563309
# 
# 	######## testing rlock managed by server process
# 
# 8192 iterations in 1.3295256 seconds
# average number/sec: 6161.59628668
# 
# 	######## testing threading.Condition
# 
# 32768 waits in 1.58998064 seconds
# average number/sec: 20609.0559694
# 
# 	######## testing multiprocessing.Condition
# 
# 65536 waits in 1.9246026 seconds
# average number/sec: 34051.7050117
# 
# 	######## testing condition managed by a server process
# 
# 8192 waits in 1.9664576 seconds
# average number/sec: 4165.86658161
# 
# 
# 
# -------------------------------------------------------------------------------------------------
# History                               
# -------------------------------------------------------------------------------------------------
# see benchmark_mp_xi2.py (SENSE), which uses standard multiprocessing

# -------------------------------------------------------------------------------------------------
# End Of test_benchmark_mp_xi2.py                               
# -------------------------------------------------------------------------------------------------